/*
 * Present how branch prediction affects program peformance.
 *
 * Intel® Core™ i5-6600 CPU @ 3.30GHz
 *
 * $ ./branch -n 16384 -t 131072 -v 0
 * Time elapsed: 6.705130 seconds.
 * $ ./branch -n 16384 -t 131072 -v 1
 * Time elapsed: 1.307566 seconds.
 * $ ./branch -n 16384 -t 131072 -v 2
 * Time elapsed: 1.756622 seconds.
 * $ ./branch -n 16384 -t 131072 -v 3
 * Time elapsed: 1.762065 seconds.
 */
#include "common.h"

/* You can modify following definitions to try out different settings. */ 
#define T short 
#define BLOCK 32

/* random order */
static __noinline void fill_1(T *arr, size_t n) {
  for (size_t i = 0; i < n; i++)
    arr[i] = (T)i - ((T)n >> 1);

  for (size_t i = 1; i < n; i++) {
    size_t j = i + fast_random() % (n - i);

    T t = arr[j]; arr[j] = arr[i]; arr[i] = t;
  }
}

/* sorted */
static __noinline void fill_2(T *arr, size_t n) {
  for (size_t i = 0; i < n; i++)
    arr[i] = (T)i - ((T)n >> 1);
}

static __noinline long walk_1(T *arr, size_t n, size_t m) {
  long r = 0;

  for (size_t j = 0; j < m; j++) {
    for (size_t i = 0; i < n; i++) {
      long k = arr[i];
      if (k < 0)
        r += - (3 * k + 3);
      else
        r += 5 * k + 5;
    }
  }

  return r;
}

static __noinline long walk_2(T *arr, size_t n, size_t m) {
  long r = 0;
  /* XXX: Fill in this procedure! */
  return r;
}

int main(int argc, char **argv) {
  int opt, variant = -1, n = 0, m = 0;
  bool err = false;

  while ((opt = getopt(argc, argv, "n:t:v:")) != -1) {
    if (opt == 'n')
      n = atoi(optarg);
    else if (opt == 't')
      m = atoi(optarg);
    else if (opt == 'v')
      variant = atoi(optarg);
    else
      err = true;
  }

  if (err || n == 0 || m == 0 || variant < 0 || variant >= 4)
    fail("Usage: %s -n size -t times -v variant\n", argv[0]);

  if (n % BLOCK)
    fail("Array size (%d) must be divisible by %d!", n, BLOCK);

  size_t size = n * sizeof(T);
  T *arr = malloc_page_aligned(size);

  printf("Generate array of %d elements (%ld KiB)\n", n, size >> 10);

  if ((variant & 1) == 0)
    fill_1(arr, n);
  else
    fill_2(arr, n);
  flush_cache();

  printf("Performing array walk %d times.\n", m);

  _timer_t timer;
  timer_reset(&timer);
  timer_start(&timer);
  long r = 0;
  if ((variant & 2) == 0)
    r = walk_1(arr, n, m);
  else
    r = walk_2(arr, n, m);
  timer_stop(&timer);
  timer_print(&timer);

  free(arr);

  printf("Result is %ld.\n", r);

  return 0;
}
