/*
 * Present how out-of-order execution affects program performance.
 *
 * Intel® Core™ i5-6600 CPU @ 3.30GHz
 *
 * $ ./combine -n 4096 -t 2097152 -v 0
 * Time elapsed: 2.874563 seconds.
 * $ ./combine -n 4096 -t 2097152 -v 1
 * Time elapsed: 2.244405 seconds.
 * $ ./combine -n 4096 -t 2097152 -v 2
 * Time elapsed: 1.880398 seconds.
 * $ ./combine -n 4096 -t 2097152 -v 3
 * Time elapsed: 1.507615 seconds.
 * $ ./combine -n 4096 -t 2097152 -v 4
 * Time elapsed: 0.999397 seconds.
 */
#include "common.h"

/* You can modify following definitions to try out different settings. */ 
#define BLOCK 8
#define IDENT 0
#define OP +
typedef long T;

static __noinline void fill(T *arr, size_t n) {
  for (size_t i = 0; i < n; i++)
    arr[i] = fast_random() | 1;
}

static __noinline T combine_1(T *arr, size_t n, size_t m) {
  T r = IDENT;
  /* XXX: Fill in this procedure! */
  return r;
} 

static __noinline T combine_2(T *arr, size_t n, size_t m) {
  T r = IDENT;
  /* XXX: Fill in this procedure! */
  return r;
} 

static __noinline T combine_3(T *arr, size_t n, size_t m) {
  T r = IDENT;
  /* XXX: Fill in this procedure! */
  return r;
} 

static __noinline T combine_4(T *arr, size_t n, size_t m) {
  T r = IDENT;
  /* XXX: Fill in this procedure! */
  return r;
} 

#pragma GCC push_options
#pragma GCC optimize ("-ftree-vectorize")
static __noinline T combine_5(T *arr, size_t n, size_t m) {
  T r = IDENT;
  /* XXX: Fill in this procedure! */
  return r;
}
#pragma GCC pop_options

int main(int argc, char **argv) {
  int opt, variant = -1, n = 0, m = 0;
  bool err = false;

  while ((opt = getopt(argc, argv, "n:t:v:")) != -1) {
    if (opt == 'n')
      n = atoi(optarg);
    else if (opt == 't')
      m = atoi(optarg);
    else if (opt == 'v')
      variant = atoi(optarg);
    else
      err = true;
  }

  if (err || n == 0 || m == 0 || variant < 0 || variant >= 5)
    fail("Usage: %s -n size -t times -v variant\n", argv[0]);

  if (n % BLOCK)
    fail("Array size (%d) must be divisible by %d!", n, BLOCK);

  size_t size = n * sizeof(T);
  T *arr = malloc_page_aligned(size);

  printf("Generate array of %d elements (%ld KiB)\n", n, size >> 10);

  fast_srandom(0xDEADC0DE);

  fill(arr, n);
  flush_cache();

  printf("Performing array walk %d times.\n", m);

  _timer_t timer;
  timer_reset(&timer);
  timer_start(&timer);
  T r = IDENT;
  if (variant == 0)
    r = combine_1(arr, n, m);
  else if (variant == 1)
    r = combine_2(arr, n, m);
  else if (variant == 2)
    r = combine_3(arr, n, m);
  else if (variant == 3)
    r = combine_4(arr, n, m);
  else
    r = combine_5(arr, n, m);
  timer_stop(&timer);
  timer_print(&timer);

  free(arr);

  printf("Result is %g\n", (double)r);

  return 0;
}
