#ifndef __COMMON_H__
#define __COMMON_H__

#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>

#define CONCAT2(A, B) A ## B
#define CONCAT(A, B) CONCAT2(A, B)

/* Block size for Intel® x86-64. */
#define BLOCK_SIZE 64
/* Hopefully no one has more than 16MiB of L3 cache. */
#define CACHE_SIZE (1 << 24)

/* Number of items of type T in array of size S. */
#define NITEMS(S, T) ((S) / sizeof(T))
/* Function attribute that prevents compiler from inlining it. */
#ifndef __noinline
#define __noinline __attribute__ ((noinline))
#endif
/* Function argument attribute to mark unused argument. */
#ifndef __unused
#define __unused __attribute__ ((unused))
#endif
/* Prevents _compiler_ from reordering memory read & write operations. */
#define barrier() asm volatile("" ::: "memory")
/* Prevents _hardware_ from reordering memory read & write operations. */
#define mfence() asm volatile("mfence")

/* Tell branch predictor that following condition is likely / unlikely
 * to be satisfied. */
#define likely(x) __builtin_expect((x), 1)
#define unlikely(x) __builtin_expect((x), 0)

typedef struct _timer {
  struct timespec start, sum;
} _timer_t;

void fail(const char *fmt, ...);

void fast_srandom(uint64_t seed);
uint64_t fast_random();

void flush_cache();

void timer_reset(_timer_t *t);
void timer_start(_timer_t *t);
void timer_stop(_timer_t *t);
void timer_print(_timer_t *t);

void read_bytes(const char *pathname, void *buf, size_t n);
void *malloc_page_aligned(size_t size);

#endif
