/*
 * row-major vs. tiled texture queries
 *
 * Intel® Core™ i5-6600 CPU @ 3.30GHz
 *
 * $ ./texture -t 65536 -v 0
 * Time elapsed: 1.700562 seconds.
 * $ ./texture -t 65536 -v 1
 * Time elapsed: 1.031514 seconds.
 */
#include "common.h"

/* You can modify following definitions to try out different settings. */ 
#define N 8192
typedef unsigned char T;

static inline T get_1(T *arr, long x, long y) {
  return arr[y * N + x];
}

static inline void put_1(T *arr, long x, long y, T value) {
  arr[y * N + x] = value;
}

#define VARIANT 1
#include "texture.h"

static inline T get_2(T *arr, long x, long y) {
  /* XXX: Fill in this procedure! */
  return 0;
}

static inline void put_2(T *arr, long x, long y, T value) {
  /* XXX: Fill in this procedure! */
}

#define VARIANT 2
#include "texture.h"

int main(int argc, char **argv) {
  int opt, variant = -1, t = 0;
  bool err = false;

  while ((opt = getopt(argc, argv, "t:v:")) != -1) {
    if (opt == 't')
      t = atoi(optarg);
    else if (opt == 'v')
      variant = atoi(optarg);
    else
      err = true;
  }

  if (err || t == 0 || variant < 0 || variant >= 2)
    fail("Usage: %s -t times -v variant\n", argv[0]);

  size_t size = N * N * sizeof(T);
  T *arr = malloc_page_aligned(size);

  printf("Generate texture of %dx%d pixels (%ld KiB)\n", N, N, size >> 10);

  fast_srandom(0xDEADC0DE);

  if (variant == 0)
    fill_1(arr);
  else
    fill_2(arr);
  flush_cache();

  printf("Performing texture queries %d times.\n", t);

  _timer_t timer;
  timer_reset(&timer);
  timer_start(&timer);
  long r = 0;
  for (int i = 0; i < t; i++) {
    long a = fast_random();
    long xs = (short)(a >> 48) & (N - 1);
    long ys = (short)(a >> 32) & (N - 1);
    long xe = (short)(a >> 16) & (N - 1);
    long ye = (short)(a) & (N - 1);
    if (variant == 0)
      r += query_1(arr, xs, ys, xe, ye);
    else
      r += query_2(arr, xs, ys, xe, ye);
  }
  timer_stop(&timer);
  timer_print(&timer);

  free(arr);

  printf("Result is %lx\n", r);

  return 0;
}
